"""
Copyright (C) 2021-2023 Julian Metzler

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import csv
import itertools


def high16(value):
    # Get high byte of a 16-bit value
    return value >> 8

def low16(value):
    # Get low byte of a 16-bit value
    return value & 0xFF


def _debug_print(debug, *args, **kwargs):
    if debug:
        print(*args, **kwargs)

def debug_hex(message, readable_ascii = False, readable_ctrl = False):
    """
    Turn a message into a readable form
    """

    CTRL_CHARS = {
        0x02: "STX",
        0x03: "ETX",
        0x04: "EOT",
        0x05: "ENQ",
        0x10: "DLE",
        0x15: "NAK",
        0x17: "ETB"
    }

    result = []
    for byte in message:
        if readable_ctrl and byte in CTRL_CHARS:
            result.append(CTRL_CHARS[byte])
        elif readable_ascii and byte not in range(0, 32) and byte != 127:
            result.append(chr(byte))
        else:
            result.append("{:02X}".format(byte))
    return " ".join(result)


def vias_in_route(route, vias):
    # Check if the given vias are all present in the given route in the right order
    # If an entry in vias is a list, all of its items will be considered to be aliases of each other
    if not vias:
        return False
    i = 0
    j = 0
    while i < len(route) and j < len(vias):
        if type(vias[j]) in (tuple, list):
            for alias in vias[j]:
                if route[i] == alias:
                    j += 1
                    break
        else:
            if route[i] == vias[j]:
                j += 1
        i += 1
    return j == len(vias)


def get_vias(route, weights, *via_groups, check_dashes=True, debug=False):
    # Get the ideal combination of vias based on split-flap modules
    num_groups = len(via_groups)
    
    # Go through all via groups and take note of possible candidates
    via_candidates = []
    for group in via_groups:
        group_candidates = []
        for pos, entry in group.items():
            if vias_in_route(route, entry['stations']):
                group_candidates.append(pos)
        via_candidates.append(group_candidates)
    _debug_print(debug, "Via candidates:")
    _debug_print(debug, via_candidates)
    
    # Check all combinations to see if the order makes sense
    combinations = itertools.product(*via_candidates)
    valid_combinations = []
    _debug_print(debug, "\nVia candidates with sensible order:")
    for combination in combinations:
        stations = []
        for group, pos in enumerate(combination):
            stations.extend(via_groups[group][pos]['stations'])
        if vias_in_route(route, stations):
            _debug_print(debug, combination, stations)
            valid_combinations.append(combination)
    
    # If check_dashes is True, check if the starts and endings are compatible,
    # i.e. if the first segment ends on a dash, the next one
    # cannot start with one.
    if check_dashes:
        valid_dash_combinations = []
        _debug_print(debug, "\nCandidates after check_dashes:")
        for combination in valid_combinations:
            valid = True
            prev_text = None
            for group, pos in enumerate(combination):
                text = via_groups[group][pos]['text'].strip()
                if group > 0:
                    if prev_text and text and prev_text.endswith("-") == text.startswith("-"):
                        _debug_print(debug, "Excluded: ", prev_text, text)
                        valid = False
                        break
                prev_text = text
            if valid:
                _debug_print(debug, combination)
                valid_dash_combinations.append(combination)
        valid_combinations = valid_dash_combinations

    # Build the texts of all valid combinations
    # and remove combinations that contain double entries
    final_combinations = []
    for combination in valid_combinations:
        text_stations = []
        for group, pos in enumerate(combination):
            text_stations.extend([s.strip() for s in via_groups[group][pos]['text'].split(" - ") if s.strip()])
        if len(set(text_stations)) == len(text_stations):
            # No double entries detected
            final_combinations.append([combination, text_stations])
    
    # Calculate the total weight of each combinations
    for i, entry in enumerate(final_combinations):
        combination, text_stations = entry
        weight = 0
        for text_station in text_stations:
            weight += weights.get(text_station, 1)
        final_combinations[i].append(weight)
    final_combinations.sort(key=lambda c: c[2], reverse=True)

    _debug_print(debug, "\nFinal combinations (Score, Positions, Text):")
    for entry in final_combinations:
        _debug_print(debug, entry[2], entry[0], " - ".join(entry[1]))
    _debug_print(debug, "")
    
    if final_combinations:
        return final_combinations[0][0]
    else:
        return None

def vias_from_csv(filename):
    # Build the dict required for get_vias from a CSV file
    vias = {}
    with open(filename, newline='', encoding='utf-8') as f:
        reader = csv.reader(f, delimiter=';', quotechar='"')
        for i, row in enumerate(reader):
            if i == 0 or not row[1]:
                continue
            vias[int(row[0])] = {
                'text': row[1],
                'stations': [[subentry.strip() for subentry in entry.split(",")] for entry in row[2:] if entry]
            }
    return vias

def map_from_csv(filename):
    # Build the dict required for SplitFlapDisplay from a CSV file.
    # CSV format: column 0 = flap position, column 1 = destination as printed on the flap
    _map = {}
    with open(filename, newline='', encoding='utf-8') as f:
        reader = csv.reader(f, delimiter=';', quotechar='"')
        for i, row in enumerate(reader):
            if i == 0 or not row[1]:
                continue
            _map[int(row[0])] = row[1]
    return _map

def alternatives_map_from_csv(filename):
    # Build the dict required for an alternative station name mapping from a CSV file.
    # CSV format: column 0 = flap position (irrelevant), column 1 = destination as printed on the flap,
    # column 2 = comma separated list of alternative station names that map to this flap
    _map = {}
    with open(filename, newline='', encoding='utf-8') as f:
        reader = csv.reader(f, delimiter=';', quotechar='"')
        for i, row in enumerate(reader):
            if i == 0 or not row[1]:
                continue
            for station_name in row[2].split(","):
                if station_name.strip() and station_name.strip() != row[1]:
                    _map[station_name.strip()] = row[1]
    return _map



"""
Copyright (C) 2020-2023 Julian Metzler

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import serial
import time

from .mis1_protocol import MIS1Protocol

from ..utils import debug_hex
from ..utils.base_serial import BaseSerialPort


class MIS1TextDisplay(MIS1Protocol):
    ALIGN_LEFT = 0x00
    ALIGN_RIGHT = 0x01
    ALIGN_CENTER = 0x02
    ALIGN_SCROLL = 0x03 # apparently not supported
    
    ATTR_BLINK = 0x01
    ATTR_INVERT = 0x02
    ATTR_BLINK_INV = 0x08
    
    DATE_FORMAT_DISABLE = 0x00
    DATE_FORMAT_DDMMYY = 0x01
    DATE_FORMAT_MMDDYY = 0x02
    DATE_FORMAT_YYMMDD = 0x03
    
    TIME_FORMAT_DISABLE = 0x00
    TIME_FORMAT_24H = 0x01
    TIME_FORMAT_12H_AM_PM = 0x02
    TIME_FORMAT_12H = 0x03
    
    def merge_attributes(self, text):
        if type(text) in (tuple, list):
            merged = ""
            for t, attrs in text:
                merged += "\x00" + chr(attrs) + t
            return merged
        return text

    def simple_text(self, page, row, col, text, align = ALIGN_LEFT):
        text = self.merge_attributes(text)
        text = text.encode("CP437")
        data = [align, page, row, col] + list(text)
        return self.send_command(0x11, 0x00, data, expect_response=False)

    def text(self, page, row, col_start, col_end, text, align = ALIGN_LEFT):
        text = self.merge_attributes(text)
        text = text.encode("CP437")
        data = [align, page, row, col_start >> 8, col_start & 0xFF, col_end >> 8, col_end & 0xFF] + list(text)
        return self.send_command(0x15, 0x00, data, expect_response=False)
    
    def set_pages(self, pages):
        flat_pages = [item for sublist in pages for item in sublist]
        data = [0x00] + flat_pages
        return self.send_command(0x24, 0x00, data, expect_response=False)
    
    def set_page(self, page):
        return self.set_pages([(page, 255)])

    def reset(self):
        return self.send_command(0x31, 0x00, [], expect_response=False)

    def set_test_mode(self, state):
        return self.send_command(0x32, 0x00, [1 if state else 0])

    def sync(self):
        return self.send_command(0x34, 0x00, [], expect_response=False)
    
    def set_clock(self, year, month, day, hour, minute, second):
        # apparently unsupported
        data = list(divmod(second, 10)[::-1])
        data += divmod(minute, 10)[::-1]
        data += divmod(hour, 10)[::-1]
        data += divmod(day, 10)[::-1]
        data += divmod(month, 10)[::-1]
        data += divmod(year, 10)[::-1]
        data += [0x00] # weekday, unused
        return self.send_command(0x3A, 0x00, data, expect_response=False)
    
    def set_clock_display(self, date_format, date_row, date_col_start, date_col_end, time_format, time_row, time_col_start, time_col_end):
        # apparently unsupported
        data = [date_format, date_row]
        data += [date_col_start >> 8, date_col_start & 0xFF]
        data += [date_col_end >> 8, date_col_end & 0xFF]
        data += [time_format, time_row]
        data += [time_col_start >> 8, time_col_start & 0xFF]
        data += [time_col_end >> 8, time_col_end & 0xFF]
        return self.send_command(0x3D, 0x00, data, expect_response=False)
    
    def set_outputs(self, states):
        # states: array of 8 bools representing outputs 0 through 7
        state_byte = 0x00
        for i in range(max(8, len(states))):
            if states[i]:
                state_byte |= (1 << i)
        return self.send_command(0x41, 0x00, [0x00, 0x00, state_byte], expect_response=False)



"""
Copyright (C) 2023 Julian Metzler

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import serial


class AesysDSA:
    def __init__(self, port, exclusive=True, debug=False):
        self.port = port
        self.debug = debug
        self.exclusive = exclusive
        self.open()
    
    def open(self):
        self.device = serial.Serial(self.port,
            baudrate=9600, bytesize=8, parity='N', stopbits=1, exclusive=self.exclusive)
    
    def close(self):
        self.device.close()
    
    def _checksum(self, data):
        checksum = sum(data)
        data += "{:04X}".format(checksum & 0xFFFF).encode('ascii')
        return data
    
    def send_text(self, text):
        data = "\x01\x17P000060{text}".format(text=text)
        length = len(data)
        frame = "\x02AVIS{length:04X}{data}\x03".format(length=length, data=data)
        frame = frame.encode('cp850')
        frame = self._checksum(frame)
        if self.debug:
            print(frame)
        self.device.write(frame)


"""
Copyright (C) 2022 Julian Metzler

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import datetime
import requests

from dateutil import parser as dt_parser


class FraportAPI:
    def __init__(self):
        pass
    
    def _convert_flight_dict(self, flight, is_arrival):
        """
        Convert a flight data dict from the Fraport API format to our format.
        This mostly means renaming keys.
        
        is_arrival: Used to rename arrival/departure time keys
        """
        scheduled_arrival_key = 'sched' if is_arrival else 'schedArr'
        scheduled_departure_key = 'schedDep' if is_arrival else 'sched'
        estimated_arrival_key = 'esti' if is_arrival else 'estiArr'
        estimated_departure_key = 'estiDep' if is_arrival else 'esti'
        
        out = {
            'aircraft_icao': flight.get('ac'),
            'aircraft_registration': flight.get('reg'),
            'airline_iata': flight.get('al'),
            'airline_name': flight.get('alname'),
            'airport_iata': flight.get('iata'),
            'airport_name': flight.get('apname'),
            'baggage_claims': flight.get('bag'),
            'codeshares': flight.get('cs'),
            'counters': flight.get('schalter'),
            'duration': datetime.timedelta(minutes=flight.get('duration')) if flight.get('duration') else None,
            'estimated_arrival': dt_parser.parse(flight.get(estimated_arrival_key)) if flight.get(estimated_arrival_key) else None,
            'estimated_departure': dt_parser.parse(flight.get(estimated_departure_key)) if flight.get(estimated_departure_key) else None,
            'exit': flight.get('ausgang'),
            'flight_id': flight.get('id'),
            'flight_number': flight.get('fnr'),
            'flight_status': flight.get('flstatus'), # unclear
            'gate': flight.get('gate'),
            'hall': flight.get('halle'),
            'language': flight.get('lang'),
            'last_update': dt_parser.parse(flight.get('lu')) if flight.get('lu') else None,
            's': flight.get('s'), # unclear
            'scheduled_arrival': dt_parser.parse(flight.get(scheduled_arrival_key)) if flight.get(scheduled_arrival_key) else None,
            'scheduled_departure': dt_parser.parse(flight.get(scheduled_departure_key)) if flight.get(scheduled_departure_key) else None,
            'status': flight.get('status'),
            'stops': flight.get('stops'),
            'terminal': flight.get('terminal'),
            'type': flight.get('typ'), # unclear
            'via_iata': flight.get('rou'),
            'via_name': flight.get('rouname')
        }
        return out
    
    def get_flights(self, flight_type='departures', count=10, lang="en", page=1, timestamp=None):
        """
        Get flight data
        
        flight_type: departures or arrivals
        count: How many flights to retrieve
        lang: Language for status fields in response
        page: which page of the results to get
        """
        get_params = {
            'perpage': count,
            'lang': lang,
            'page': page,
            'flighttype': flight_type,
            'time': (timestamp or datetime.datetime.utcnow()).strftime('%Y-%m-%dT%H:%M:00.000Z')
        }
        is_arrival = flight_type == 'arrivals'
        resp = requests.get("https://www.frankfurt-airport.com/de/_jcr_content.flights.json/filter", params=get_params)
        data = resp.json()
        #with open("out.json", 'r') as f:
        #    data = json.load(f)
        #with open("out.json", 'w') as f:
        #    json.dump(data, f)
        data['flights'] = [self._convert_flight_dict(flight, is_arrival=is_arrival) for flight in data['data']]
        del data['data']
        return data



"""
Copyright (C) 2020-2023 Julian Metzler

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import serial
import time

from .exceptions import CommunicationError, DisplayError
from ..utils import debug_hex
from ..utils.base_serial import BaseSerialPort


class MIS1Protocol:
    ERROR_CODES = {
        0:  "OK",
        1:  "EFONTSIZE",
        2:  "CLIPPED",
        5:  "ECHAR_ATTRIBUTE",
        6:  "EINIT_DISPLAY",
        8:  "ERANGE",
        11: "EOPTION",
        13: "EHWDISPLAY",
        15: "ESECTOR",
        16: "EPAGE",
        17: "BITMAP",
        19: "EBITMAP",
        20: "ELINE",
        24: "ESECTORNR",
        25: "EPAGENR",
        26: "ESECX",
        27: "ESECY",
        36: "EFONTNR",
        38: "EFONTMISS",
        57: "EALLOC",
        255: "EGENERAL"
    }

    def __init__(self, port, address=1, baudrate=9600, exclusive=True, debug=False, rx_timeout=3.0, use_rts=True):
        self.address = address
        self.debug = debug
        self.rx_timeout = rx_timeout
        self.use_rts = use_rts
        if isinstance(port, serial.Serial) or isinstance(port, BaseSerialPort):
            self.port = port
        else:
            self.port = serial.Serial(port, baudrate=baudrate, bytesize=8, parity="E", stopbits=1, exclusive=exclusive, timeout=rx_timeout)

    def checksum(self, data):
        checksum = 0x00
        for i, byte in enumerate(data):
            checksum += byte
        return (checksum % 256) | 0x80

    def escape(self, data):
        escaped = []
        for byte in data:
            if byte in (0x02, 0x03, 0x04, 0x05, 0x10):
                escaped += [0x10, byte]
            else:
                escaped.append(byte)
        return escaped

    def unescape(self, data):
        unescaped = []
        escape_active = False
        for byte in data:
            if not escape_active and byte == 0x10:
                escape_active = True
            else:
                escape_active = False
                unescaped.append(byte)
        return unescaped
    
    def send_raw_data(self, data):
        if self.debug:
            print("TX: " + debug_hex(data, readable_ascii=False, readable_ctrl=False))
        if self.use_rts:
            self.port.setRTS(1)
        self.port.write(data)
        if self.use_rts:
            time.sleep(0.1)
            self.port.setRTS(0)
    
    def send_raw_telegram(self, data):
        telegram = [0x04, (0x80 | self.address), 0x02] + self.escape(data) + [0x03] + [self.checksum(data + [0x03])]
        self.send_raw_data(telegram)
    
    def send_command(self, code, subcode, data, expect_response=True):
        self.send_raw_telegram([code, subcode] + data)
        if expect_response:
            return self.read_response(ack=False)

    def send_tx_request(self):
        self.send_raw_data([0x04, (0x80 | self.address), 0x05])
        return self.read_response(ack=True)

    def read_response(self, ack=False):
        data = self.port.read(1)
        if not data:
            raise CommunicationError("No response received from display")

        start = data[0]
        if start in (0x06, 0x15):
            response = data
        else:
            while start != 0x02:
                data = self.port.read(1)
                if not data:
                    raise CommunicationError("No response received from display")
                start = data[0]

            response = [start]
            escape_active = False
            while True:
                data = self.port.read(1)
                if not data:
                    raise CommunicationError("No response received from display")
                byte = data[0]
                if not escape_active:
                    if byte == 0x10:
                        escape_active = True
                    elif byte == 0x03:
                        response.append(byte)
                        data = self.port.read(1)
                        if not data:
                            raise CommunicationError("No response received from display")
                        checksum = data[0]
                        response.append(checksum)
                        break
                    else:
                        response.append(byte)
                else:
                    escape_active = False
                    response.append(byte)

        if self.debug:
            print("RX: " + debug_hex(response, readable_ascii=False, readable_ctrl=False))

        # Send ACK if required and didn't get a NAK
        if ack and response != b'\x15':
            self.send_raw_data([0x06])
        return response

    def check_error(self, response):
        if len(response) < 7:
            return
        if response[1] != 0x8C:
            return
        data = self.unescape(response[3:-2])
        raise DisplayError(self.ERROR_CODES.get(data[1], str(data[1])))



"""
Copyright (C) 2023 Julian Metzler

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import base64
import requests

from ..splitflap_display import SplitFlapDisplay, TextField, CustomMapField


class xatLabsCheetah:
    """
    Controls the xatLabs Cheetah universal display controller.
    This uses TCP-based protocol in which the framebuffer is
    transmitted as a Base64 encoded string.
    The reason it's done this was is to allow transferring the
    framebuffer inside a JSON object.
    Cheetah supports three types of displays:
    - Pixel-based displays (not yet implemented here)
    - Character-based displays (e.g. a character LCD or alphanumerical split-flap display)
    - Selection-based displays (e.g. a rolling film or generic split-flap display)
    """

    def __init__(self, host):
        self.host = host
        self.display_info = None
        self.device_info = None
        self.load_display_info()
        self.load_device_info()
        self.init_framebuf()

    def init_framebuf(self):
        display_type = self.display_info.get('type')
        if display_type == 'character':
            self.framebuf = [0] * self.display_info["charbuf_size"]
        elif display_type == 'selection':
            self.framebuf = [0] * self.display_info["framebuf_size"]
            if 'units' in self.display_info['config']:
                for unit in self.display_info['config']['units']:
                    if 'addr' in unit and 'home' in unit:
                        self.framebuf[unit['addr']] = unit['home']
        else:
            raise NotImplementedError("Unsupported display type: {}".format(display_type))

    def get_actual_char_count(self, s):
        count = 0
        if "combining_full_stop" in self.display_info["quirks"]:
            for i in range(len(s)):
                if i == 0:
                    count += 1  # Always count first character
                elif s[i] == ".":
                    if s[i-1] == ".":
                        count += 1  # Only count full stop if preceded by another full stop
                else:
                    count += 1  # Always count non full stop characters
        else:
            count = len(s)
        return count

    def update_framebuf_text(self, text):
        self.init_framebuf()
        val = ""
        for e in text.split("\n"):
            actual_len = self.get_actual_char_count(e)
            excess = actual_len - self.display_info["width"]
            if excess < 0:
                # Need to pad
                val += e.ljust(len(e) - excess, " ")
            else:
                # Need to cut off
                cutoff_pos = len(e) - excess
                val += e[:cutoff_pos]
        val_len = len(val)
        for i in range(self.display_info["charbuf_size"]):
            if i >= val_len:
                self.framebuf[i] = 0
            else:
                code = ord(val[i])
                self.framebuf[i] = code if code <= 255 else 0

    def update_framebuf_sel(self, module_data):
        self.init_framebuf()
        for pos, val in module_data.items():
            self.framebuf[pos] = int(val)

    def load_display_info(self):
        resp = requests.get(f"{self.host}/info/display.json")
        self.display_info = resp.json()

    def load_device_info(self):
        resp = requests.get(f"{self.host}/info/device.json")
        self.device_info = resp.json()

    def set_text(self, text):
        if self.display_info.get('type') != 'character':
            raise NotImplementedError("Only character displays support set_text")
        self.update_framebuf_text(text)
        buf = bytearray(self.framebuf)
        buffer_b64 = base64.b64encode(buf).decode('ascii')
        requests.post(f"{self.host}/canvas/buffer.json", json={'buffer': buffer_b64})

    def set_brightness(self, brightness):
        if not self.display_info.get('brightness_control'):
            raise NotImplementedError("Display does not support brightness control")
        assert brightness in range(0, 256)
        requests.post(f"{self.host}/canvas/brightness.json", json={'brightness': brightness})

    def d_set_module_data(self, module_data):
        # Compatibility function for SplitFlapDisplay class
        self.update_framebuf_sel(dict(module_data))

    def d_update(self):
        # Compatibility function for SplitFlapDisplay class
        buf = bytearray(self.framebuf)
        buffer_b64 = base64.b64encode(buf).decode('ascii')
        requests.post(f"{self.host}/canvas/buffer.json", json={'buffer': buffer_b64})

    def get_splitflap_display(self):
        if self.display_info.get('type') != 'selection':
            raise NotImplementedError("Only selection displays support get_splitflap_display")
        unit_data = self.display_info.get('config', {}).get('units')
        if not unit_data:
            raise NotImplementedError("Display does not provide layout data")
        map_data = self.display_info.get('config', {}).get('maps')
        if not map_data:
            raise NotImplementedError("Display does not provide mapping data")

        display = SplitFlapDisplay(self)
        for unit in unit_data:
            if unit['type'] == 'map':
                field = CustomMapField(map_data.get(unit.get('map', None), {}), start_address=unit['addr'], length=unit['len'], x=unit['x'], y=unit['y'], module_width=unit['width'], module_height=unit['height'], home_pos=unit['home'])
                setattr(display, unit['name'], field)
            elif unit['type'] == 'text':
                field = TextField(start_address=unit['addr'], length=unit['len'], x=unit['x'], y=unit['y'], module_width=unit['width'], module_height=unit['height'], home_pos=unit['home'], display_mapping=map_data.get(unit.get('map', None), None))
                setattr(display, unit['name'], field)
        return display


"""
Copyright (C) 2022 Julian Metzler

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import datetime
import socket
import time

from io import BytesIO
from PIL import Image


class ECSLCDisplay:
    """
    ERROR / STATUS CODES
    """

    CMD_STATUS_DICT = {
        -1: "ERR_NO_RESPONSE",
        0:  "ECS_LCD_OK",
        1:  "ECS_LCD_EFONTSIZE",
        2:  "ECS_LCD_CLIPPED",
        3:  "ECS_LCD_ENOTFOUND",
        4:  "ECS_LCD_EDEVICECMD",
        5:  "ECS_LCD_ECHAR_ATTRIBUTE",
        6:  "ECS_LCD_EINIT_DISPLAY",
        7:  "ECS_LCD_EINIT_SECTOR",
        8:  "ECS_LCD_ERANGE",
        9:  "ECS_LCD_EOPTION_CONVERT",
        10: "ECS_LCD_EOPTION_LEN",
        11: "ECS_LCD_EOPTION",
        12: "ECS_LCD_EMISOPTION",
        13: "ECS_LCD_EHWDISPLAY",
        14: "ECS_LCD_DISPLAY",
        15: "ECS_LCD_ESECTOR",
        16: "ECS_LCD_EPAGE",
        17: "ECS_LCD_EPAGE_SEQ",
        18: "ECS_LCD_ECOMMUNICATION",
        19: "ECS_LCD_EBITMAP",
        20: "ECS_LCD_ELINE",
        21: "ECS_LCD_ESYNTAX",
        22: "ECS_LCD_EDELTEXT",
        23: "ECS_LCD_ENOPAGEACT",
        24: "ECS_LCD_ESECTORNR",
        25: "ECS_LCD_EPAGENR",
        26: "ECS_LCD_ESECX",
        27: "ECS_LCD_ESECY",
        28: "ECS_LCD_EFONT_MISS",
        29: "ECS_LCD_ESECLIN",
        30: "ECS_LCD_ENOSECACT",
        31: "ECS_LCD_EFEWPARS",
        32: "ECS_LCD_EOPTION1",
        33: "ECS_LCD_EOPTION2",
        34: "ECS_LCD_EOPTION3",
        35: "ECS_LCD_EOPTION4",
        36: "ECS_LCD_EFONTNR",
        37: "ECS_LCD_EFONTMISS",
        38: "ECS_LCD_ESECTTYPE",
        39: "ECS_LCD_EMUCHPARS",
        40: "ECS_LCD_EDEC_MISS",
        41: "ECS_LCD_COMMAND",
        42: "ECS_LCD_ENOEND",
        43: "ECS_LCD_EPREFIX",
        44: "ECS_LCD_EPOSTFIX",
        45: "ECS_LCD_EALIGN",
        46: "ECS_LCD_EATTRIB",
        47: "ECS_LCD_ECOLOR",
        48: "ECS_LCD_ESAMECOLORS",
        49: "ECS_LCD_ENOSPACE",
        50: "ECS_LCD_ENOANZPAGE",
        51: "ECS_LCD_EOUTOFRANGE",
        52: "ECS_LCD_ENOPAGENOTIME",
        53: "ECS_LCD_ECASE",
        54: "ECS_LCD_EONOFF",
        55: "ECS_LCD_EBIT",
        56: "ECS_LCD_EVALUE",
        57: "ECS_LCD_EALLOC",
        58: "ECS_LCD_EFILE",
        59: "ECS_LCD_EADRDOUBLE",
        60: "ECS_LCD_EGMFC",
        61: "ECS_LCD_EBLOCK",
        62: "ECS_LCD_EFONT",
        63: "ECS_LCD_EGLASTAB",
        64: "ECS_LCD_EHWNOTSUPPORTED",
        254: "ECS_LCD_ECONFIG",
        255: "ECS_LCD_EGENERAL"
    }

    OP_STATUS_DICT = {
        0: "OP_STAT_READY",
        1: "OP_STAT_LAMP_FAULT",
        2: "OP_STAT_LCD_FAULT",
        3: "OP_STAT_LAMP_LCD_FAULT"
    }


    def __init__(self, host, port=7487, timeout=10.0, double_sided=True, debug=False):
        self.host = host
        self.port = port
        self.timeout = timeout
        self.double_sided = double_sided
        self.debug = debug
        self.socket = None
        self.socket_rfile = None
        self.socket_wfile = None
        socket.setdefaulttimeout(timeout)
        self._init_socket()
        device_cfg = self.get_device_cfg()
        self.width = device_cfg['xPixel']
        if self.double_sided:
            # Double-sided displays have the back side mapped below the front side
            self.height = device_cfg['yPixel'] // 2
        else:
            self.height = device_cfg['yPixel']


    """
    SOCKET I/O
    """

    def _init_socket(self):
        if self.socket is not None:
            try:
                self.socket.close()
            except:
                pass
        self.socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.socket.connect((self.host, self.port))
        self.socket_rfile = self.socket.makefile('r', encoding='latin-1')
        self.socket_wfile = self.socket.makefile('w', encoding='latin-1')

    def write(self, data):
        self.socket_wfile.write(data)
        self.socket_wfile.flush()

    def read_response(self):
        try:
            response = self.socket_rfile.readline()
            if self.debug:
                print("Received response: " + response.strip())
            return response
        except (socket.timeout, OSError):
            self._init_socket()
            return None

    def parse_response(self):
        response = self.read_response()
        if response is None:
            return {'status': -1, 'status_text': self.CMD_STATUS_DICT.get(-1, "UNKNOWN")}

        # Stupid bug, some replies have a space before the equals sign
        response = response.replace(" =", "=")

        parts = response.split()
        command = parts[0][:-1] # -1 to remove the trailing colon

        resp_dict = {}
        for part in parts[1:]:
            subparts = part.split("=")
            resp_dict[subparts[0]] = "=".join(subparts[1:])

        for int_key in ('status', 'operative', 'xPixel', 'yPixel'):
            if int_key in resp_dict:
                resp_dict[int_key] = int(resp_dict[int_key])

        if 'status' in resp_dict:
            resp_dict['status_text'] = self.CMD_STATUS_DICT.get(resp_dict['status'], "UNKNOWN")
        if 'operative' in resp_dict:
            resp_dict['operative_text'] = self.OP_STATUS_DICT.get(resp_dict['operative'], "UNKNOWN")

        if command == "ECS_LCD_GetTime":
            if 'time' in resp_dict:
                resp_dict['time'] = datetime.datetime.strptime(resp_dict['time'], "%H:%M:%S").time()
            if 'date' in resp_dict:
                resp_dict['date'] = datetime.datetime.strptime(resp_dict['date'], "%Y-%m-%d").date()
            if 'time' in resp_dict and 'date' in resp_dict:
                resp_dict['datetime'] = datetime.datetime.combine(resp_dict['date'], resp_dict['time'])
        elif command == "ECS_LCD_GetResetTime":
            if 'time' in resp_dict:
                resp_dict['time'] = int(resp_dict['time'])

        return resp_dict

    def send_command(self, command, *args, **kwargs):
        kwargs_list = []
        for key, value in kwargs.items():
            if value is None:
                continue

            if type(value) in (tuple, list):
                value_str = " ".join(map(str, value))
            elif type(value) in (bytes, bytearray):
                value_str = " ".join(["{:02X}".format(b) for b in value])
            else:
                value_str = str(value)
            kwargs_list.append("-" + key + " " + value_str)

        args_str = " ".join(map(str, args))
        kwargs_str = " ".join(kwargs_list)
        command_str = command.strip()
        if args_str:
            command_str += " " + args_str
        if kwargs_str:
            command_str += " " + kwargs_str
        command_str = command_str.strip() + "\r\n"
        if self.debug:
            print("Sending command:   " + command_str.strip())
        self.write(command_str)
        return self.parse_response()


    """
    COMMANDS
    """

    def define_normal_page(self, page):
        """
        Sets up a normal page with the given page ID.
        Normal pages are exactly the same size as the display.
        
        page: page ID, 0...255
        """
        return self.send_command("ECS_LCD_DefineNormalPage", page=page)

    def define_virtual_page(self, page, cols, rows):
        """
        Sets up a virtual page with the given page ID.
        Virtual pages can be larger than the display size and support
        vertical scrolling of rows.
        
        page: page ID, 0...255
        cols: Page width in columns
        rows: Page height in rows
        """
        return self.send_command("ECS_LCD_DefineVirtualPage", page=page, size=(cols, rows))

    def delete_page(self, page):
        """
        Deletes the page with the given ID.
        
        page: page ID, 0...255
        """
        return self.send_command("ECS_LCD_DeletePage", page=page)

    def select_page(self, page):
        """
        Selects the page with the given ID for editing.
        
        page: page ID, 0...255
        """
        return self.send_command("ECS_LCD_SelectPage", page=page)

    def change_sector(self, sector, col, row, cols, rows, type, font=None, lines=None, bg_color=None):
        """
        Updates the currently selected sector on the currently
        selected page with the given configuration.
        
        sector: sector ID, 0...255
        col: sector origin column, 0...65024
        row: sector origin row, 0...65024
        cols: sector width in columns, 0...65024
        rows: sector height in rows, 0...65024 (for virtual pages, only height 1 is valid)
        type: sector type, text or graphic
        font: font ID, 0...255 (only for sector type text)
        lines: number of lines of text in the sector, 0...255 (only for sector type text)
        bg_color: background color of the sector, 3-tuple, (0, 0, 0)...(255, 255, 255) (only for displays with RGB backlight)
        """
        return self.send_command("ECS_LCD_ChangeSector", sector=(sector, col, row, cols, rows), sectortype=type, font=font, lines=lines, bkcolor=bg_color)

    def select_sector(self, sector):
        """
        Selects the sector with the given ID on the currently selected page for editing.
        
        sector: sector ID, 0...255
        """
        return self.send_command("ECS_LCD_SelectSector", sector=sector)

    def chr_out_text(self, row, text, align, attribute, textcode=None):
        """
        Outputs text in the currently selected sector on the currently selected page.
        
        row: line index inside the current sector, 0...255
        text: text to output
        align: text alignment (left, center, right, flow) (flow: scroll text if it doesn't fit into the sector)
        attribute: text display attribute (normal, blink, compblink, invers) (compblink: inverted and blinking)
        textcode: index of pre-defined text for split-flap style control, 0...65535
        """
        text = "\"" + text.replace("\"", "") + "\""
        return self.send_command("ECS_LCD_ChrOutText", row, text=text, align=align, textcode=textcode, attribute=attribute)

    def put_image(self, col, row, image, filename=None):
        """
        Outputs the given image in the currently selected sector on the currently selected page.
        
        col: X position of the image in columns, 0...65024
        row: Y position of the image in rows, 0...65024
        image: image to output (path to image file, PIL Image object, or raw BMP image data)
        filename: filename to save the image as on the display (max. 31 characters)
        """
        if isinstance(image, Image.Image):
            bio = BytesIO()
            image.convert('1')
            image.save(bio, 'BMP')
            bio.seek(0)
            img_data = bio.read()
        elif type(image) is str:
            bio = BytesIO()
            image = Image.open(image)
            image.convert('1')
            image.save(bio, 'BMP')
            bio.seek(0)
            img_data = bio.read()
        else:
            img_data = image
        return self.send_command("ECS_LCD_PutImage", col, row, imagefilename=filename, image=img_data)

    def insert_line(self, line):
        """
        Inserts a line of text into the currently selected sector,
        below the line with the specified index.
        The remaining lines will be shifted downwards.
        
        line: line ID, 0...255
        """
        return self.send_command("ECS_LCD_InsertLine", line=line)

    def delete_line(self, line):
        """
        Deletes a line of text from the currently selected sector.
        The remaining lines will be shifted upwards.
        
        line: line ID, 0...255
        """
        return self.send_command("ECS_LCD_DeleteLine", line=line)

    def chr_clr_txt(self, line):
        """
        CLears a line of text in the currently selected sector.
        
        line: line ID, 0...255
        """
        return self.send_command("ECS_LCD_ChrClrTxt", line)

    def clr_sector(self):
        """
        Deletes the contents of the currently selected sector.
        """
        return self.send_command("ECS_LCD_ClrSector")

    def clr_page(self):
        """
        Deletes the contents of the currently selected page.
        """
        return self.send_command("ECS_LCD_ClrPage")

    def show_page(self, page):
        """
        Displays the page with the given ID on the display.
        
        page: page ID, 0...255
        """
        return self.send_command("ECS_LCD_ShowPage", page=page)

    def prog_page_sequence(self, page, duration):
        """
        Appends an entry to the current list of pages to show.
        Each entry comprises the page ID to be shown and a duration.
        
        page: page ID, 0...255
        duration: duration is milliseconds, 0...65024
        """
        return self.send_command("ECS_LCD_ProgPageSequence", page=page, time=duration)

    def activate_page_sequence(self):
        """
        Activates the previously programmed page seqneuce.
        """
        return self.send_command("ECS_LCD_ActivatePageSequence")

    def scroll_page(self, mode, lines=None, pause=None):
        """
        Scrolls the display contents up by the given amount of lines of text
        with the given pause between lines.
        
        mode: scroll mode (off, line, cont)
                off:  disable scrolling
                line: scroll row by row
                cont: scroll pixel by pixel
        lines: the number of lines to scroll, 0...255
        pause: the pause between lines in seconds, 0...255
        """
        return self.send_command("ECS_LCD_ScrollPage", scroll=mode, lines=lines, pausetime=pause)

    def switch_background_light(self, state):
        """
        Turns the background light on or off.
        
        state: light state (on, off)
        """
        if type(state) in (bool, int):
            state = 'on' if state else 'off'
        return self.send_command("ECS_LCD_SwitchBackgroundLight", light=state)

    def get_operative_status(self):
        """
        Get the operative status of the display.
        """
        return self.send_command("ECS_LCD_GetOperativeStatus")

    def get_reset_time(self):
        """
        Get the number of seconds since the last reset.
        """
        return self.send_command("ECS_LCD_GetResetTime")

    def get_time(self):
        """
        Get the system time.
        """
        return self.send_command("ECS_LCD_GetTime")

    def get_software_version(self):
        """
        Get the software version.
        """
        return self.send_command("ECS_LCD_GetSoftwareVersion")

    def get_device_cfg(self):
        """
        Get the device configuration.
        """
        return self.send_command("ECS_LCD_GetDeviceCfg")

    def get_page_content(self, page):
        """
        Get the contents of the specified page.

        page: page ID, 0...255
        """
        return self.send_command("ECS_LCD_GetPageContent", page=page)


    """
    CONVENIENCE FUNCTIONS
    """

    def display_text(self, page, text, font, align='center', attribute='normal'):
        """
        Displays a text on the given page.

        page: page ID, 0...255
        text: the text to display
        font: font ID, 0...255
        align: text alignment (left, center, right, flow)
        attribute: text display attribute (normal, blink, compblink, invers) (compblink: inverted and blinking)
        """
        lines = text.splitlines()
        self.define_normal_page(page)
        self.select_page(page)
        self.change_sector(0, 0, 0, self.width, self.height, 'text', font, lines=len(lines))
        self.select_sector(0)
        for i, row in enumerate(lines):
            self.chr_out_text(i, row, align, attribute)

        if self.double_sided:
            self.change_sector(1, 0, self.height, self.width, self.height, 'text', font, lines=len(lines))
            self.select_sector(1)
            for i, row in enumerate(lines):
                self.chr_out_text(i, row, align, attribute)
        
        self.show_page(page)

    def display_image(self, page, image, halign='center', valign='middle'):
        """
        Displays an image on the given page.

        page: page ID, 0...255
        image: the image to display (PIL Image object or path to image)
        halign: horizontal image alignment (left, center, right)
        valign: vertical image alignment (top, middle, bottom)
        """
        if not isinstance(image, Image.Image):
            image = Image.open(image)
        width, height = image.size

        if halign == 'center':
            x = (self.width - width) // 2
        elif halign == 'right':
            x = self.width - width
        else:
            x = 0

        if valign == 'middle':
            y = (self.height - height) // 2
        elif valign == 'bottom':
            y = self.height - height
        else:
            y = 0

        self.define_normal_page(page)
        self.select_page(page)
        self.change_sector(0, 0, 0, self.width, self.height, 'graphic')
        self.select_sector(0)
        self.put_image(x, y, image)

        if self.double_sided:
            self.change_sector(1, 0, self.height, self.width, self.height, 'graphic')
            self.select_sector(1)
            self.put_image(x, y, image)
        
        self.show_page(page)


"""
Copyright (C) 2023 Julian Metzler

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import json
import os


def get_text_width(text, font):
    """
    Get the width of the given text using the given font.
    """
    
    if not text:
        return 0
    
    try:
        with open(os.path.join(os.path.dirname(__file__), "dimensions-{}.json".format(font)), 'r') as f:
            dimensions = json.load(f)
        width = 0
        for char in text:
            dims = dimensions.get(char, (0, 0))
            w = dims[0]
            if w is None:
                w = 0
            width += w
        width += dimensions.get('spacing', 0) * (len(text) - 1)
        return width
    except FileNotFoundError:
        raise NotImplementedError("Width calculation not available for this font")


"""
Copyright (C) 2022 Julian Metzler

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import math
import serial
import time

from .mis1_text import MIS1TextDisplay
from ..utils.base_serial import BaseSerialPort


class MIS1Board:
    """
    A large board using multiple GCUs
    """
    
    def __init__(self, port, start_address = 1, num_rows = 1, rows_per_gcu = 8, baudrate = 9600, exclusive = True, debug = False):
        """
        port: Serial port for the GCU bus
        start_address: Address of the first GCU, usually 1
        num_rows: How many rows of text the board has
        rows_per_gcu: How many rows are controlled per GCU
        baudrate: GCU baudrate
        exclusive: Whether to lock the serial port for exclusive access
        debug: Enable debug output
        """
        self.start_address = start_address
        self.debug = debug
        self.num_rows = num_rows
        self.rows_per_gcu = rows_per_gcu
        
        if isinstance(port, serial.Serial) or isinstance(port, BaseSerialPort):
            self.port = port
        else:
            self.port = serial.Serial(port, baudrate=baudrate, bytesize=8, parity="E", stopbits=1, exclusive=exclusive)
            
        self.num_gcus = math.ceil(num_rows / rows_per_gcu)
        self.gcus = []
        self.gcu_outputs = []
        for i in range(self.start_address, self.start_address + self.num_gcus):
            self.gcus.append(MIS1GCUDisplay(self.port, i, debug=self.debug))
            self.gcu_outputs.append([0] * 8)
    
    def write_row(self, page, row, col, text):
        """
        Write text on specified page, row and column of the board (both starting at 0)
        """
        gcu_index = row // self.rows_per_gcu
        gcu_row = row % self.rows_per_gcu
        self.gcus[gcu_index].simple_text(page, gcu_row, col, text)
    
    def show_page(self, page):
        """
        Show the given page
        """
        for gcu in self.gcus:
            gcu.set_page(page)
    
    def show_pages(self, pages):
        """
        Show the given pages. Structure:
        [(page_num, page_duration), (page_num, page_duration)]
        where page_duration is as follows:
        0 = 0.0 s (invalid)
        1 = 0.5 s
        2 = 1.0 s
        ...
        254 = 127.0 s
        255 = 127.5 s (maximum)
        """
        for gcu in self.gcus:
            gcu.set_pages(pages)
    
    def write_text(self, page, start_row, start_col, text):
        """
        Write the given multiline text to the board on the given page
        and starting at the given row and column
        """
        lines = text.splitlines()
        for row in range(start_row, min(self.num_rows, start_row + len(lines))):
            self.write_row(page, row, start_col, lines[row - start_row])
    
    def set_blinker(self, row, state):
        """
        On boards with blinkers (like airport info boards), set the blinker
        for the given row active or inactive. This assumes the blinkers are
        controlled by the GPIOs of the GCU in ascending order.
        """
        gcu_index = row // self.rows_per_gcu
        gcu_row = row % self.rows_per_gcu
        self.gcu_outputs[gcu_index][gcu_row] = 1 if state else 0
    
    def update_blinkers(self):
        """
        Update the GPIOs according to the internal blinker states
        """
        for i, gcu in enumerate(self.gcus):
            gcu.set_outputs(self.gcu_outputs[i])



"""
Copyright (C) 2022 Julian Metzler

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

HAS_TKINTER = False
try:
    import tkinter as tk
    import tkinter.font as tkFont
    HAS_TKINTER = True
except ModuleNotFoundError:
    pass

from .fields import *


"""
This module can build a GUI based on a SplitFlapDisplay instance.
This can be used to easily build a manual control tool for a display.
"""

class SplitFlapGUI:
    def __init__(self, display, parent):
        if not HAS_TKINTER:
            raise RuntimeError("Tkinter could not be loaded. Please make sure the tkinter module exists.")
        
        self.display = display
        self.parent = parent
        self.field_widgets = {}
        self.frame = tk.Frame(parent)
        self.build_gui()

    @staticmethod
    def _set_optionmenu_width(widget, choices):
        f = tkFont.nametofont(widget.cget("font"))
        zerowidth = f.measure("0")
        w = round(max([f.measure(i) for i in choices]) / zerowidth)
        widget.config(width=w)

    @staticmethod
    def _validate_entry(allowed_chars, max_length, text):
        if len(text) > max_length:
            return False
        if allowed_chars:
            for char in text:
                if char not in (allowed_chars + [" "]):
                    return False
        return True

    def build_gui(self):
        fields = self.display.get_fields()
        for name, field in fields:
            if isinstance(field, TextField):
                if field.display_mapping:
                    allowed_chars = list(set([c.lower() for c in field.display_mapping.values()] + [c.upper() for c in field.display_mapping.values()]))
                else:
                    allowed_chars = None
                _len = field.length # Required! If you use field.length in the lambda directly, it'll give the wrong value!
                _validate_func = lambda text: self._validate_entry(allowed_chars, _len, text)
                _vcmd = (self.frame.register(_validate_func), "%P")
                entry = tk.Entry(self.frame, validate='key', validatecommand=_vcmd)
                entry.grid(column=field.x, row=field.y, columnspan=field.module_width*field.length, rowspan=field.module_height, sticky=tk.NSEW, padx=2, pady=2)
                self.field_widgets[name] = (entry, None, None)
            elif isinstance(field, CustomMapField):
                choices = list(field.display_mapping.values())
                var = tk.StringVar(self.frame)
                var.set(choices[0])
                opt = tk.OptionMenu(self.frame, var, *choices)
                opt.configure(indicatoron=False, anchor=tk.W)
                self._set_optionmenu_width(opt, choices)
                opt.grid(column=field.x, row=field.y, columnspan=field.module_width, rowspan=field.module_height, sticky=tk.NSEW, padx=2, pady=2)
                self.field_widgets[name] = (opt, var, choices)

    def update_display(self):
        for name, widget_data in self.field_widgets.items():
            widget, var, choices = widget_data
            if var:
                getattr(self.display, name).set(var.get())
            else:
                getattr(self.display, name).set(widget.get())
        #self.display.update()

    def clear_display(self):
        self.display.clear()
        for name, widget_data in self.field_widgets.items():
            widget, var, choices = widget_data
            if choices:
                var.set("")
            else:
                widget.delete(0, tk.END)
        #self.display.update()


"""
Copyright 2019 - 2023 Julian Metzler

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import serial
import time

from .exceptions import CommunicationError
from ..utils.base_serial import BaseSerialPort


class Krone8200PST:
    """
    Controls the PST bus in a Krone 8200 split-flap display.
    """

    def __init__(self, port, debug = False, exclusive = True):
        self.debug = debug
        if isinstance(port, serial.Serial) or isinstance(port, BaseSerialPort):
            self.port = port
        else:
            self.port = serial.Serial(port, baudrate=2400, timeout=1.0, exclusive=exclusive)

    def debug_message(self, message):
        """
        Turn a message into a readable form
        """
        result = ""
        for byte in message:
            result += "{:02X} ".format(byte)
        return result

    def send_raw_message(self, message):
        if self.debug:
            print("TX: " + self.debug_message(message))
        for byte in message:
            self.port.write(bytearray([byte]))
            time.sleep(0.05)

    def set_home(self):
        """
        Set all units to their home position
        """
        return self.send_raw_message([0x1B])

    def set_unit(self, address, position):
        """
        Set a given unit to a given position
        """
        return self.send_raw_message([0x3A, address, position])

    def update(self):
        """
        Update units (cause them to actually turn)
        """
        return self.send_raw_message([0x1C])

    def reset(self):
        """
        Reset all unit controllers
        """
        return self.send_raw_message([0x1A])

    def d_set_module_data(self, module_data):
        # Compatibility function for SplitFlapDisplay class
        for addr, pos in module_data:
            self.set_unit(addr, pos)

    def d_update(self):
        # Compatibility function for SplitFlapDisplay class
        self.update()



"""
Copyright (C) 2022 Julian Metzler

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import socket
from .base_serial import BaseSerialPort


class TcpSerialPort(BaseSerialPort):
    def __init__(self, host, port, timeout=2.0):
        """
        host: The hostname or IP to connect to
        port: The TCP port to use for communication
        timeout: The socket timeout in seconds
        """
        self.socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.socket.connect((host, port))
        self.socket.settimeout(timeout)

    def write(self, data):
        return self.socket.send(bytearray(data))

    def read(self, length):
        # Read the specified number of bytes, blocking
        return self.socket.recv(length)

    def setRTS(self, state):
        pass

    def setDTR(self, state):
        pass

    def getCTS(self):
        return 0

    def getDSR(self):
        return 0

    def getRI(self):
        return 0

    def getCD(self):
        return 0

    def __del__(self):
        self.socket.close()



"""
Copyright (C) 2023 Julian Metzler

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import serial
import time

from PIL import Image

from .mis1_protocol import MIS1Protocol

from ..utils import debug_hex, high16, low16
from ..utils.base_serial import BaseSerialPort


class MIS1MatrixDisplay(MIS1Protocol):
    def set_config(self, lcd_module, num_lcds, x, y, id, board_timeout, fr_freq, fps, is_master, protocol_timeout, response_delay):
        # board_timeout: in seconds
        # protocol_timeout and response_delay: in milliseconds
        data = [
            high16(lcd_module),
            low16(lcd_module),
            num_lcds,
            high16(x),
            low16(x),
            high16(y),
            low16(y),
            high16(id),
            low16(id),
            high16(board_timeout * 2),
            low16(board_timeout * 2),
            fr_freq,
            fps,
            1 if is_master else 0,
            high16(protocol_timeout),
            low16(protocol_timeout),
            high16(response_delay),
            low16(response_delay),
            0x00,
            0x00
        ]
        return self.send_command(0x02, 0x00, data)

    def set_input_config(self, id, mask_byte, row_codes):
        data = [id, mask_byte]
        data.extend(row_codes)
        return self.send_command(0x04, 0x00, data)

    def text(self, board, page, font, x, y, width, text):
        data = [
            board,
            page,
            font,
            high16(x),
            low16(x),
            high16(y),
            low16(y),
            high16(width),
            low16(width)
        ]
        data.extend(bytearray(text, 'ascii'))
        return self.send_command(0x19, 0x00, data)

    def check_text_width(self, font, text):
        data = [0x00, font]
        data.extend(bytearray(text, 'ascii'))
        return self.send_command(0x1A, 0x00, data)

    def set_font_spacing(self, font_spacings):
        data = font_spacings
        return self.send_command(0x1F, 0x00, data)
    
    def set_pages(self, pages):
        flat_pages = [item for sublist in pages for item in sublist]
        data = [0x00, 0x00]
        data.extend(flat_pages)
        return self.send_command(0x24, 0x00, data)
    
    def set_page(self, page):
        return self.set_pages([(page, 255)])

    def copy_page(self, source, destination):
        data = [0x00, source, destination]
        return self.send_command(0x26, 0x00, data)

    def echo(self, number):
        data = [high16(number), low16(number)]
        return self.send_command(0x2A, 0x00, data)

    def set_pixel(self, page, x, y, state):
        data = [
            0x00,
            page,
            high16(x),
            low16(x),
            high16(y),
            low16(y),
            1 if state else 0
        ]
        return self.send_command(0x2B, 0x00, data)

    def fill_area(self, page, x, y, width, height, state):
        data = [
            0x00,
            page,
            high16(x),
            low16(x),
            high16(y),
            low16(y),
            high16(width),
            low16(width),
            high16(height),
            low16(height),
            1 if state else 0
        ]
        return self.send_command(0x2D, 0x00, data)

    def delete_scroll_sector(self, sector, page):
        data = [0x00, sector, page]
        return self.send_command(0x2E, 0x00, data)

    def delete_page(self, page):
        data = [0x00, page]
        return self.send_command(0x2F, 0x00, data)

    def reset(self):
        return self.send_command(0x31, 0x00, [])

    def set_test_mode(self, state):
        return self.send_command(0x32, 0x00, [1 if state else 0])

    def sync(self):
        return self.send_command(0x34, 0x00, [])

    def get_defective_rows(self):
        return self.send_command(0x35, 0x00, [])

    def get_config(self):
        return self.send_command(0x38, 0x00, [])

    def get_font_info(self):
        return self.send_command(0x39, 0x00, [])

    def partial_page_update(self, page, x, y, height, width):
        data = [
            0x00,
            0x00,
            page,
            high16(x),
            low16(x),
            high16(y),
            low16(y),
            high16(height),
            low16(height),
            high16(width),
            low16(width)
        ]
        return self.send_command(0x3A, 0x00, data)

    def get_id(self):
        return self.send_command(0x3E, 0x00, [])
    
    def set_outputs(self, states):
        # states: array of 8 bools representing outputs 0 through 7
        state_byte = 0x00
        for i in range(max(8, len(states))):
            if states[i]:
                state_byte |= (1 << i)
        return self.send_command(0x41, 0x00, [0x00, 0x00, state_byte])

    def read_inputs(self, id):
        data = [id]
        return self.send_command(0x42, 0x00, data)

    def set_vlcd(self, voltage):
        ones = int(voltage)
        decimals = round((voltage % 1) * 100)
        data = [ones, decimals]
        return self.send_command(0x61, 0x00, data)

    def get_firmware_revision(self):
        return self.send_command(0x65, 0x00, [])

    def get_temperature(self):
        return self.send_command(0x6A, 0x00, [])

    def get_vlcd(self):
        return self.send_command(0x6E, 0x00, [])

    def image_data(self, page, x, y, width, pixels):
        data = [
            0x00,
            0x00,
            page,
            high16(x),
            low16(x),
            high16(y),
            low16(y),
            high16(width),
            low16(width)
        ]
        data.extend(pixels)
        return self.send_command(0x73, 0x00, data)

    def scroll_image_data(self, sector, page, y, pixels):
        data = [
            0x00,
            0x00,
            sector,
            page,
            high16(y),
            low16(y)
        ]
        data.extend(pixels)
        return self.send_command(0x74, 0x00, data)

    def create_scroll_area(self, sector, page, x, y, width, height, data_width):
        data = [
            0x00,
            0x00,
            sector,
            page,
            high16(x),
            low16(x),
            high16(y),
            low16(y),
            high16(width),
            low16(width),
            high16(height),
            low16(height),
            high16(data_width),
            low16(data_width)
        ]
        return self.send_command(0x75, 0x00, data)

    def set_flash_cycle(self, cycle_time):
        data = [cycle_time]
        return self.send_command(0x7C, 0x00, data)

    def become_slave(self):
        return self.send_command(0x7D, 0x00, [])

    def become_master(self):
        return self.send_command(0x7E, 0x00, [])

    def dummy(self):
        return self.send_command(0x7F, 0x00, [])

    def image(self, page, x, y, image):
        if not isinstance(image, Image.Image):
            image = Image.open(image)
        image = image.convert('L')
        pixels = image.load()
        width, height = image.size
        for y_offset in range(height):
            _y = y + y_offset
            pixel_data = []
            byte = 0x00
            x_bit = 7
            for x_offset in range(width):
                if pixels[x_offset, y_offset] > 127:
                    byte |= (1 << x_bit)
                if x_bit == 0 or x_offset == width - 1:
                    x_bit = 7
                    pixel_data.append(byte)
                    byte = 0x00
                else:
                    x_bit -= 1
            self.image_data(page, x, _y, width, pixel_data)

    def scroll_image(self, sector, page, x, y, scroll_width, image, extra_whitespace=0):
        if not isinstance(image, Image.Image):
            image = Image.open(image)
        image = image.convert('L')
        pixels = image.load()
        width, height = image.size
        
        self.create_scroll_area(sector, page, x, y, scroll_width, height, width + extra_whitespace)
        for i in range(10):
            response = self.send_tx_request()
            self.check_error(response)

        for y_offset in range(height):
            _y = y + y_offset
            pixel_data = []
            byte = 0x00
            x_bit = 7
            for x_offset in range(width):
                if pixels[x_offset, y_offset] > 127:
                    byte |= (1 << x_bit)
                if x_bit == 0 or x_offset == width - 1:
                    x_bit = 7
                    pixel_data.append(byte)
                    byte = 0x00
                else:
                    x_bit -= 1
            self.scroll_image_data(sector, page, y_offset, pixel_data)

    def animation(self, sector_start, page, x, y, image):
        # Splits the animation into 1-pixel wide scroll sectors
        # that scroll through a spatial representation of the animation
        if not isinstance(image, Image.Image):
            image = Image.open(image)
        width, height = image.size

        orig_frames = []
        try:
            while True:
                orig_frames.append(image.convert('L').load())
                # Next frame
                image.seek(image.tell() + 1)
        except EOFError:
            pass

        num_frames = len(orig_frames)
        scroll_frames = []
        for i in range(width):
            scroll_frame = Image.new('L', (num_frames, height), 'black')
            scroll_pixels = scroll_frame.load()
            for _x, frame in enumerate(orig_frames):
                for _y in range(height):
                    scroll_pixels[_x, _y] = frame[i, _y]
            scroll_frames.append(scroll_frame)

        for _x, scroll_frame in enumerate(scroll_frames):
            self.scroll_image(sector_start + _x, page, x + _x, y, 1, scroll_frame)



"""
Copyright (C) 2023 Julian Metzler

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import crcmod
import socket
import time

from PIL import Image, ImageOps, ImageSequence


class VistraI:
    """
    A VISTRA-I display
    
    Please note that currently only Ethernet connection is supported.
    """
    
    MSG_TYPE_CLEAR_PANEL = 0x01
    MSG_TYPE_NULL = 0x06
    MSG_TYPE_SET_BRIGHTNESS = 0x08
    MSG_TYPE_TEXT = 0x0B
    MSG_TYPE_CLEAR_WINDOW = 0x0C
    MSG_TYPE_GRAPHICS = 0x0E
    MSG_TYPE_GET_BITMAP = 0xF2
    MSG_TYPE_GET_TEXT_OBJECTS = 0xF5
    
    EFFECT_NONE = 0x00
    EFFECT_BLINK = 0x01
    EFFECT_SCROLL_ALTERNATE = 0x02
    EFFECT_SCROLL = 0x04
    EFFECT_SCROLL_VERTICAL = 0x08
    EFFECT_CENTERED = 0x10
    EFFECT_RIGHT = 0x20
    EFFECT_MIDDLE = 0x40
    EFFECT_BOTTOM = 0x80

    EFFECT3_NONE = 0x00
    EFFECT3_DEF = 0x03 # purpose unclear
    EFFECT3_WINDOW_INVERTED = 0x40
    EFFECT3_INVERTED = 0x50
    
    def __init__(self, hostname, port, address = 0, timeout = 5.0):
        """
        hostname:
        The network hostname of the display to connect to
        
        port:
        The TCP port to use for communication
        
        address:
        The panel address (0 for all panels, only useful for RS232)
        """
        
        self.hostname = hostname
        self.port = port
        self.address = address
        self.timeout = timeout
        self.crc = crcmod.predefined.mkPredefinedCrcFun('crc-ccitt-false')
        self.queue = None
        self.socket = None
        self.last_transmission = 0
        #self.renew_socket()
    
    def renew_socket(self):
        """
        Renew the socket in case it got fucked up
        """
        
        try:
            self.socket.close()
        except:
            pass
        self.socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.socket.settimeout(self.timeout)
        self.socket.connect((self.hostname, self.port))
        #print("Socket renewed.")
    
    def init_queue(self):
        """
        Start collecting messages for bulk sending
        instead of directly sending them.
        """
        
        self.queue = []
    
    def send_queue(self):
        """
        Send all collected messages in bulk.
        """
        
        queue = self.queue
        self.queue = None
        reply = self.send_message(queue)
        return reply
    
    def wrap_message(self, message):
        """
        Add the start bytes and end byte to a message.
        
        message:
        The message as a bytestring
        """
        
        message = bytearray([0x02, self.address]) + message
        message += bytearray([0x03])
        return message
    
    def wrap_partial_message(self, message):
        """
        Add the start byte and CRC to a message.
        
        message:
        The message as a bytestring
        """
        
        crc = self.crc(message)
        crc_msb = crc >> 8
        crc_lsb = crc & 0xFF
        message = bytearray([0x1B]) + message
        message += bytearray([crc_lsb, crc_msb])
        return message
    
    def send_raw_message(self, message):
        """
        Send a raw message to the display.
        
        message:
        The message as a bytestring
        """
        
        try:
            #print(message)
            self.socket.send(message)
            # Receive up until the "RequestedDataType" byte
            reply = bytearray(self.socket.recv(9))
            #print(reply)
            data_type = reply[8]
            if data_type != 0x00:
                # Receive data length
                reply += bytearray(self.socket.recv(4))
                length = (
                    (reply[12] << 24)
                    | (reply[11] << 16)
                    | (reply[10] << 8)
                    | reply[9])
                # Receive the rest of the data
                reply += bytearray(self.socket.recv(length+1)) # +1 for end byte
            else:
                # Receive end byte
                reply += bytearray(self.socket.recv(1))
            self.last_transmission = time.time()
            return reply
        except socket.timeout:
            # Silently try to renew socket and fail silently
            try:
                self.renew_socket()
            except:
                pass
            raise
    
    def send_message(self, message):
        """
        Send a message to the display OR queue it if the queue mode is enabled.
        
        message:
        The message as a dict of attributes OR a list of message dicts.
        Example:
            {
                'msgType': 0...255,
                'attr2': 0...255,
                'attr3': 0...255,
                'font': 0...255,
                'winX': 0...65535,
                'winY': 0...65535,
                'winWidth': 0...65535,
                'winHeight': 0...65535,
                'data': bytearray([0x00, 0xFF])
            }
        """
        
        if type(message) not in (list, tuple):
            message = [message]
        
        if self.queue is not None:
            self.queue += message
            return
        
        complete_message = bytearray()
        for m in message:
            msg = bytearray([
                m.get('msgType', self.MSG_TYPE_TEXT),
                m.get('attr2', 0),
                m.get('attr3', 0),
                m.get('font', 0),
                (m.get('winX', 0) & (2**16-1)) & 0xFF, # support for negative values
                (m.get('winX', 0) & (2**16-1)) >> 8,
                (m.get('winY', 0) & (2**16-1)) & 0xFF,
                (m.get('winY', 0) & (2**16-1)) >> 8,
                (m.get('winWidth', 0) & (2**16-1)) & 0xFF,
                (m.get('winWidth', 0) & (2**16-1)) >> 8,
                (m.get('winHeight', 0) & (2**16-1)) & 0xFF,
                (m.get('winHeight', 0) & (2**16-1)) >> 8,
                len(m.get('data', bytearray())) & 0xFF,
                len(m.get('data', bytearray())) >> 8
            ])
            data = m.get('data', bytearray())
            if type(data) is str:
                data = data.encode('latin-1', 'replace')
            msg += bytearray(data)
            msg = self.wrap_partial_message(msg)
            complete_message += msg
        
        complete_message = self.wrap_message(complete_message)
        # Renew socket if necessary
        if time.time() - self.last_transmission > 300: # 5 minutes
            self.renew_socket()
        return self.send_raw_message(complete_message)
    
    def clear_panel(self):
        """
        Clear all panel contents.
        """
        
        message = {'msgType': self.MSG_TYPE_CLEAR_PANEL}
        return self.send_message(message)
    
    def null_cmd(self):
        """
        Send a null command to keep communication going.
        """
        
        message = {'msgType': self.MSG_TYPE_NULL}
        return self.send_message(message)
    
    def set_brightness(self, brightness):
        """
        Set the base display brightness.
        
        brightness:
        The brightness level (0...255)
        """
        
        message = {
            'msgType': self.MSG_TYPE_SET_BRIGHTNESS,
            'data': [brightness]
        }
        return self.send_message(message)
    
    def clear_window(self, x = 0, y = 0):
        """
        Clear a window specified by the x and y coordinates.
        
        x, y:
        The coordinates of the window to be cleared
        """
        
        message = {
            'msgType': self.MSG_TYPE_CLEAR_WINDOW,
            'winX': x,
            'winY': y
        }
        return self.send_message(message)
    
    def send_text(self, text = "", font = 0,
        x = 0, y = 0, width = 0, height = 0,
        effects = EFFECT_NONE, effects3 = EFFECT3_NONE):
        """
        Send a text to the display.
        
        text:
        The text to send
        
        font:
        The ID of the font to use
        
        x, y:
        x and y coordinates of the top left window corner
        
        width, height:
        Window size
        
        effects:
        The effects to apply to the text (OR of the effect flags defined above)
        
        effects3:
        Some further effects
        """
        
        message = {
            'msgType': self.MSG_TYPE_TEXT,
            'attr2': effects,
            'attr3': effects3,
            'font': font,
            'winX': x,
            'winY': y,
            'winWidth': width,
            'winHeight': height,
            'data': text
        }
        return self.send_message(message)
    
    def send_image(self, image, invert = False,
        x = 0, y = 0, width = 0, height = 0,
        effects = EFFECT_NONE, effects3 = EFFECT3_NONE):
        """
        Send an image to the display.
        
        image:
        The image to send (file path, open file or Image instance)
        
        invert:
        Whether to invert the colors before sending
        
        x, y:
        x and y coordinates of the top left window corner
        
        width, height:
        Window size
        
        effects:
        The effects to apply to the image (OR of the effect flags defined above)
        
        effects3:
        Some further effects
        """
        
        if not isinstance(image, Image.Image):
            image = Image.open(image)
        
        image = image.convert("L")
        if invert:
            image = ImageOps.invert(image)
        im_width, im_height = image.size
        pixels = image.load()
        
        data = [
            im_width & 0xFF,
            im_width >> 8,
            im_height & 0xFF,
            im_height >> 8,
            0x00,
            0x00
        ]
        
        for im_y in range(0, im_height, 8):
            for im_x in range(im_width):
                byte = 0x00
                for im_yoff in range(7, -1, -1):
                    try:
                        byte |= (pixels[im_x, im_y+im_yoff] > 127) << im_yoff
                    except IndexError:
                        byte |= 1 << im_yoff
                data.append(byte)
        
        message = {
            'msgType': self.MSG_TYPE_GRAPHICS,
            'attr2': effects,
            'attr3': effects3,
            'winX': x,
            'winY': y,
            'winWidth': width,
            'winHeight': height,
            'data': data
        }
        return self.send_message(message)
    
    def send_gif(self, image, delay = 0.1, *args, **kwargs):
        """
        Send an animated GIF (one cycle).
        
        image:
        The image to send
        """
        
        if not isinstance(image, Image.Image):
            img = Image.open(image)
        else:
            img = image
        
        for frame in ImageSequence.Iterator(img):
            self.send_image(frame, *args, **kwargs)
            time.sleep(delay)
    
    def get_bitmap(self):
        """
        Get the current display bitmap.
        
        Returns:
        A PIL Image representing the current bitmap
        """
        
        message = {
            'msgType': self.MSG_TYPE_GET_BITMAP
        }
        response = self.send_message(message)
        img_info = response[13:19]
        img_data = response[19:-1]
        width = (img_info[1] << 8) | img_info[0]
        height = (img_info[3] << 8) | img_info[2]
        img = Image.new("RGB", (width, height))
        pixels = img.load()
        for i, byte in enumerate(img_data):
            x = i%width
            y_base = i//width * 8
            for n in range(8):
                y = y_base + n
                active = bool((byte >> n) & 0x01)
                pixels[x, y] = (255, 215, 0) if active else (0, 0, 0)
        return img
    
    #TODO
    def get_text_objects(self):
        """
        Get the current text objects.
        
        Returns:
        ???
        """
        
        message = {
            'msgType': self.MSG_TYPE_GET_TEXT_OBJECTS
        }
        response = self.send_message(message)
        return response